package envoy.client.data;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;

import envoy.data.ConfigItem;
import envoy.data.IDGenerator;
import envoy.util.SerializationUtils;

/**
 * Implements a {@link LocalDB} in a way that stores all information inside a
 * folder on the local file system.<br>
 * <br>
 * Project: <strong>envoy-client</strong><br>
 * File: <strong>PersistentLocalDB.java</strong><br>
 * Created: <strong>27.10.2019</strong><br>
 *
 * @author Kai S. K. Engelbart
 * @author Maximilian K&auml;fer
 * @since Envoy Client v0.1-alpha
 */
public class PersistentLocalDB extends LocalDB {

	private File localDBDir, localDBFile, usersFile, idGeneratorFile, messageCacheFile, statusCacheFile;

	/**
	 * Initializes an empty local database without a directory. All changes made to
	 * this instance cannot be saved to the file system.<br>
	 * <br>
	 * This constructor shall be used in conjunction with the {@code ignoreLocalDB}
	 * {@link ConfigItem}.
	 *
	 * @since Envoy Client v0.3-alpha
	 */
	public PersistentLocalDB() {}

	/**
	 * Constructs an empty local database. To serialize any chats to the file
	 * system, call {@link PersistentLocalDB#initializeUserStorage()}.
	 *
	 * @param localDBDir the directory in which to store users and chats
	 * @throws IOException if the PersistentLocalDB could not be initialized
	 * @since Envoy Client v0.1-alpha
	 */
	public PersistentLocalDB(File localDBDir) throws IOException {
		this.localDBDir = localDBDir;

		// Initialize local database directory
		if (localDBDir.exists() && !localDBDir.isDirectory())
			throw new IOException(String.format("LocalDBDir '%s' is not a directory!", localDBDir.getAbsolutePath()));
		usersFile		= new File(localDBDir, "users.db");
		idGeneratorFile	= new File(localDBDir, "id_generator.db");
	}

	/**
	 * Creates a database file for a user-specific list of chats.<br>
	 * {@inheritDoc}
	 *
	 * @throws NullPointerException if the client user is not yet specified
	 * @since Envoy Client v0.1-alpha
	 */
	@Override
	public void initializeUserStorage() {
		if (user == null) throw new NullPointerException("Client user is null");
		localDBFile			= new File(localDBDir, user.getID() + ".db");
		messageCacheFile	= new File(localDBDir, user.getID() + "_message_cache.db");
		statusCacheFile		= new File(localDBDir, user.getID() + "_status_cache.db");
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void save() throws IOException {
		// Save users
		SerializationUtils.write(usersFile, users);

		// Save user data
		if (user != null) {
			SerializationUtils.write(localDBFile, chats);
			SerializationUtils.write(messageCacheFile, messageCache);
			SerializationUtils.write(statusCacheFile, statusCache);
		}

		// Save id generator
		if (hasIDGenerator()) SerializationUtils.write(idGeneratorFile, idGenerator);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void loadUsers() throws ClassNotFoundException, IOException { users = SerializationUtils.read(usersFile, HashMap.class); }

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void loadUserData() throws ClassNotFoundException, IOException {
		chats			= SerializationUtils.read(localDBFile, ArrayList.class);
		messageCache	= SerializationUtils.read(messageCacheFile, Cache.class);
		statusCache		= SerializationUtils.read(statusCacheFile, Cache.class);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void loadIDGenerator() {
		try {
			idGenerator = SerializationUtils.read(idGeneratorFile, IDGenerator.class);
		} catch (ClassNotFoundException | IOException e) {}
	}
}
