package envoy.client.data;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import java.util.prefs.Preferences;

import envoy.util.SerializationUtils;

/**
 * Manages all application settings, which are different objects that can be
 * changed during runtime and serialized them by using either the file system or
 * the {@link Preferences} API.<br>
 * <br>
 * Project: <strong>envoy-client</strong><br>
 * File: <strong>Settings.java</strong><br>
 * Created: <strong>11 Nov 2019</strong><br>
 *
 * @author Leon Hofmeister
 * @author Maximilian K&auml;fer
 * @author Kai S. K. Engelbart
 * @since Envoy Client v0.2-alpha
 */
public class Settings {

	// Actual settings accessible by the rest of the application
	private Map<String, SettingsItem<?>>	items;

	/**
	 * Settings are stored in this file.
	 */
	private static final File settingsFile = new File(ClientConfig.getInstance().getHomeDirectory(), "settings.ser");

	/**
	 * Singleton instance of this class.
	 */
	private static Settings settings = new Settings();

	/**
	 * The way to instantiate the settings. Is set to private to deny other
	 * instances of that object.
	 *
	 * @since Envoy Client v0.2-alpha
	 */
	private Settings() {
		// Load settings from settings file
		try {
			items = SerializationUtils.read(settingsFile, HashMap.class);
		} catch (ClassNotFoundException | IOException e) {
			items = new HashMap<>();
		}
		supplementDefaults();
	}

	/**
	 * This method is used to ensure that there is only one instance of Settings.
	 *
	 * @return the instance of Settings
	 * @since Envoy Client v0.2-alpha
	 */
	public static Settings getInstance() { return settings; }

	/**
	 * Updates the preferences when the save button is clicked.
	 *
	 * @throws IOException if an error occurs while saving the themes
	 * @since Envoy Client v0.2-alpha
	 */
	public void save() throws IOException {

		// Save settings to settings file
		SerializationUtils.write(settingsFile, items);
	}

	private void supplementDefaults() {
		items.putIfAbsent("enterToSend", new SettingsItem<>(true, "Enter to send", "Sends a message by pressing the enter key."));
		items.putIfAbsent("onCloseMode", new SettingsItem<>(true, "Hide on close", "Hides the chat window when it is closed."));
		items.putIfAbsent("currentTheme", new SettingsItem<>("dark", "Current Theme Name", "The name of the currently selected theme."));
	}

	/**
	 * @return the name of the currently active theme
	 * @since Envoy Client v0.2-alpha
	 */
	public String getCurrentTheme() { return (String) items.get("currentTheme").get(); }

	/**
	 * Sets the name of the current theme.
	 *
	 * @param themeName the name to set
	 * @since Envoy Client v0.2-alpha
	 */
	public void setCurrentTheme(String themeName) { ((SettingsItem<String>) items.get("currentTheme")).set(themeName); }

	/**
	 * @return {@code true}, if pressing the {@code Enter} key suffices to send a
	 *         message. Otherwise it has to be pressed in conjunction with the
	 *         {@code Control} key.
	 * @since Envoy Client v0.2-alpha
	 */
	public Boolean isEnterToSend() { return (Boolean) items.get("enterToSend").get(); }

	/**
	 * Changes the keystrokes performed by the user to send a message.
	 *
	 * @param enterToSend If set to {@code true} a message can be sent by pressing
	 *                    the {@code Enter} key. Otherwise it has to be pressed in
	 *                    conjunction with the {@code Control} key.
	 * @since Envoy Client v0.2-alpha
	 */
	public void setEnterToSend(boolean enterToSend) { ((SettingsItem<Boolean>) items.get("enterToSend")).set(enterToSend); }

	/**
	 * @return the current on close mode.
	 * @since Envoy Client v0.3-alpha
	 */
	public Boolean getCurrentOnCloseMode() { return (Boolean) items.get("onCloseMode").get(); }

	/**
	 * Sets the current on close mode.
	 *
	 * @param currentOnCloseMode the on close mode that should be set.
	 * @since Envoy Client v0.3-alpha
	 */
	public void setCurrentOnCloseMode(boolean currentOnCloseMode) { ((SettingsItem<Boolean>) items.get("onCloseMode")).set(currentOnCloseMode); }

	/**
	 * @return the items
	 */
	public Map<String, SettingsItem<?>> getItems() { return items; }

	/**
	 * @param items the items to set
	 */
	public void setItems(Map<String, SettingsItem<?>> items) { this.items = items; }
}
